@@ -1,4 +1,4 @@
-use proc_macro2::{Delimiter, TokenStream, TokenTree};
+use proc_macro2::{Delimiter, Ident, TokenStream, TokenTree};
 use std::iter;
 use syn::{Error, Lit, LitBool, LitStr, Result};
 
@@ -30,6 +30,16 @@ where
 	}
 }
 
+pub(crate) trait IntoIdent {
+	fn into_ident(self) -> Ident;
+}
+
+impl IntoIdent for LitStr {
+	fn into_ident(self) -> Ident {
+		Ident::new(&self.value(), self.span())
+	}
+}
+
 pub(crate) trait ExpectLit {
 	fn expect_bool(self) -> Result<LitBool>;
 	fn expect_str(self) -> Result<LitStr>;
